* {
  font-family: Arial, Helvetica, sans-serif;
  scrollbar-width: none;
  -ms-overflow-style: none;
}
* ::-webkit-scrollbar {
  display: none;
}

body {
  margin: 0 1rem;
  border: 0;
  padding: 0;
}

.code {
  font-family: "Courier New", Courier, monospace;
}

.container {
  margin: 0;
  border: 0;
  padding: 0;
  overflow: scroll;
}

.inline-info {
  word-break: break-all;
}

#main-container {
  margin: 1rem auto;
}

#main-table {
  margin-inline: auto;
  text-align: center;
}
#main-table th,
#main-table td {
  padding: 0.5rem 1rem;
}

/*# sourceMappingURL=style.css.map */
